"""The `pydantic_extra_types.json_schema` module provides the
[`JsonSchema`][pydantic_extra_types.json_schema.JsonSchema] data type.
"""

from __future__ import annotations

import json
from copy import deepcopy
from typing import Any

from pydantic import GetCoreSchemaHandler
from pydantic_core import PydanticCustomError, core_schema

try:
    from jsonschema import Draft202012Validator
    from jsonschema.exceptions import SchemaError
except ModuleNotFoundError as e:  # pragma: no cover
    raise RuntimeError(
        'The `json_schema` module requires "jsonschema" to be installed. You can install it with "pip install jsonschema".'
    ) from e


class JsonSchema(dict[str, Any]):
    """A JSON Schema validated via [`jsonschema`](https://pypi.org/project/jsonschema/).

    ```py
    from pydantic import BaseModel

    from pydantic_extra_types.json_schema import JsonSchema


    class ToolDefinition(BaseModel):
        input_schema: JsonSchema


    tool = ToolDefinition(input_schema='{"type": "object", "properties": {"name": {"type": "string"}}}')
    print(tool.input_schema)
    # > {'type': 'object', 'properties': {'name': {'type': 'string'}}}
    ```
    """

    @classmethod
    def _validate(cls, __input_value: str | dict[str, Any], _: core_schema.ValidationInfo) -> JsonSchema:
        if isinstance(__input_value, str):
            try:
                schema = json.loads(__input_value)
            except json.JSONDecodeError as exc:
                raise PydanticCustomError('json_schema_invalid_json', 'Input string must be valid JSON') from exc
        else:
            schema = __input_value

        if not isinstance(schema, dict):
            raise PydanticCustomError('json_schema_type', 'JSON Schema must be a JSON object')

        try:
            Draft202012Validator.check_schema(schema)
        except SchemaError as exc:
            raise PydanticCustomError(
                'json_schema_invalid',
                'Input must be a valid JSON Schema: {reason}',
                {'reason': exc.message},
            ) from exc

        return cls(deepcopy(schema))

    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        source: type[Any],
        handler: GetCoreSchemaHandler,
    ) -> core_schema.CoreSchema:
        return core_schema.with_info_after_validator_function(
            cls._validate,
            core_schema.union_schema(
                [
                    core_schema.str_schema(),
                    core_schema.dict_schema(
                        keys_schema=core_schema.str_schema(strict=True),
                        values_schema=core_schema.any_schema(),
                    ),
                ]
            ),
        )



"""Native Pendulum DateTime object implementation. This is a copy of the Pendulum DateTime object, but with a Pydantic
CoreSchema implementation. This allows Pydantic to validate the DateTime object.
"""

from __future__ import annotations

try:
    from pendulum import Date as _Date
    from pendulum import DateTime as _DateTime
    from pendulum import Duration as _Duration
    from pendulum import Interval as _Interval
    from pendulum import Time as _Time
    from pendulum import parse
except ModuleNotFoundError as e:  # pragma: no cover
    raise RuntimeError(
        'The `pendulum_dt` module requires "pendulum" to be installed. You can install it with "pip install pendulum".'
    ) from e
from datetime import date, datetime, time, timedelta
from typing import Any

from pydantic import GetCoreSchemaHandler
from pydantic_core import PydanticCustomError, core_schema


class DateTimeSettings(type):
    def __new__(cls, name, bases, dct, **kwargs):  # type: ignore[no-untyped-def]
        dct['strict'] = kwargs.pop('strict', True)
        return super().__new__(cls, name, bases, dct)

    def __init__(cls, name, bases, dct, **kwargs):  # type: ignore[no-untyped-def]
        super().__init__(name, bases, dct)
        cls.strict = kwargs.get('strict', True)


class DateTime(_DateTime, metaclass=DateTimeSettings):
    """A `pendulum.DateTime` object. At runtime, this type decomposes into pendulum.DateTime automatically.
    This type exists because Pydantic throws a fit on unknown types.

    ```python
    from pydantic import BaseModel
    from pydantic_extra_types.pendulum_dt import DateTime


    class test_model(BaseModel):
        dt: DateTime


    print(test_model(dt='2021-01-01T00:00:00+00:00'))

    # > test_model(dt=DateTime(2021, 1, 1, 0, 0, 0, tzinfo=FixedTimezone(0, name="+00:00")))
    ```
    """

    __slots__: list[str] = []

    @classmethod
    def __get_pydantic_core_schema__(cls, source: type[Any], handler: GetCoreSchemaHandler) -> core_schema.CoreSchema:
        """Return a Pydantic CoreSchema with the Datetime validation

        Args:
            source: The source type to be converted.
            handler: The handler to get the CoreSchema.

        Returns:
            A Pydantic CoreSchema with the Datetime validation.
        """
        return core_schema.no_info_wrap_validator_function(cls._validate, core_schema.datetime_schema())

    @classmethod
    def _validate(cls, value: Any, handler: core_schema.ValidatorFunctionWrapHandler) -> DateTime:
        """Validate the datetime object and return it.

        Args:
            value: The value to validate.
            handler: The handler to get the CoreSchema.

        Returns:
            The validated value or raises a PydanticCustomError.
        """
        # if we are passed an existing instance, pass it straight through.
        if isinstance(value, (_DateTime, datetime)):
            return DateTime.instance(value)
        try:
            # probably the best way to have feature parity with
            # https://docs.pydantic.dev/latest/api/standard_library_types/#datetimedatetime
            value = handler(value)
            return DateTime.instance(value)
        except ValueError:
            try:
                value = parse(value, strict=cls.strict)
                if isinstance(value, _DateTime):
                    return DateTime.instance(value)
                raise ValueError(f'value is not a valid datetime it is a {type(value)}')
            except ValueError:
                raise
            except Exception as exc:
                raise PydanticCustomError('value_error', 'value is not a valid datetime') from exc


class Time(_Time):
    """A `pendulum.Time` object. At runtime, this type decomposes into pendulum.Time automatically.
    This type exists because Pydantic throws a fit on unknown types.

    ```python
    from pydantic import BaseModel
    from pydantic_extra_types.pendulum_dt import Time


    class test_model(BaseModel):
        dt: Time


    print(test_model(dt='00:00:00'))

    # > test_model(dt=Time(0, 0, 0))
    ```
    """

    __slots__: list[str] = []

    @classmethod
    def __get_pydantic_core_schema__(cls, source: type[Any], handler: GetCoreSchemaHandler) -> core_schema.CoreSchema:
        """Return a Pydantic CoreSchema with the Time validation

        Args:
            source: The source type to be converted.
            handler: The handler to get the CoreSchema.

        Returns:
            A Pydantic CoreSchema with the Time validation.
        """
        return core_schema.no_info_wrap_validator_function(cls._validate, core_schema.time_schema())

    @classmethod
    def _validate(cls, value: Any, handler: core_schema.ValidatorFunctionWrapHandler) -> Time:
        """Validate the Time object and return it.

        Args:
            value: The value to validate.
            handler: The handler to get the CoreSchema.

        Returns:
            The validated value or raises a PydanticCustomError.
        """
        # if we are passed an existing instance, pass it straight through.
        if isinstance(value, (_Time, time)):
            return Time.instance(value, tz=value.tzinfo)

        # otherwise, parse it.
        try:
            parsed = parse(value, exact=True)
            if isinstance(parsed, _DateTime):
                dt = DateTime.instance(parsed)
                return Time.instance(dt.time())
            if isinstance(parsed, _Time):
                return Time.instance(parsed)
            raise ValueError(f'value is not a valid time it is a {type(parsed)}')
        except Exception as exc:
            raise PydanticCustomError('value_error', 'value is not a valid time') from exc


class Date(_Date):
    """A `pendulum.Date` object. At runtime, this type decomposes into pendulum.Date automatically.
    This type exists because Pydantic throws a fit on unknown types.

    ```python
    from pydantic import BaseModel
    from pydantic_extra_types.pendulum_dt import Date


    class test_model(BaseModel):
        dt: Date


    print(test_model(dt='2021-01-01'))

    # > test_model(dt=Date(2021, 1, 1))
    ```
    """

    __slots__: list[str] = []

    @classmethod
    def __get_pydantic_core_schema__(cls, source: type[Any], handler: GetCoreSchemaHandler) -> core_schema.CoreSchema:
        """Return a Pydantic CoreSchema with the Date validation

        Args:
            source: The source type to be converted.
            handler: The handler to get the CoreSchema.

        Returns:
            A Pydantic CoreSchema with the Date validation.
        """
        return core_schema.no_info_wrap_validator_function(cls._validate, core_schema.date_schema())

    @classmethod
    def _validate(cls, value: Any, handler: core_schema.ValidatorFunctionWrapHandler) -> Date:
        """Validate the date object and return it.

        Args:
            value: The value to validate.
            handler: The handler to get the CoreSchema.

        Returns:
            The validated value or raises a PydanticCustomError.
        """
        # if we are passed an existing instance, pass it straight through.
        if isinstance(value, (_Date, date)):
            return Date(value.year, value.month, value.day)

        # otherwise, parse it.
        try:
            parsed = parse(value)
            if isinstance(parsed, (_DateTime, _Date)):
                return Date(parsed.year, parsed.month, parsed.day)
            raise ValueError(f'value is not a valid date it is a {type(parsed)}')
        except Exception as exc:
            raise PydanticCustomError('value_error', 'value is not a valid date') from exc


class Duration(_Duration):
    """A `pendulum.Duration` object. At runtime, this type decomposes into pendulum.Duration automatically.
    This type exists because Pydantic throws a fit on unknown types.

    ```python
    from pydantic import BaseModel
    from pydantic_extra_types.pendulum_dt import Duration


    class test_model(BaseModel):
        delta_t: Duration


    print(test_model(delta_t='P1DT25H'))

    # > test_model(delta_t=Duration(days=2, hours=1))
    ```
    """

    __slots__: list[str] = []

    @classmethod
    def __get_pydantic_core_schema__(cls, source: type[Any], handler: GetCoreSchemaHandler) -> core_schema.CoreSchema:
        """Return a Pydantic CoreSchema with the Duration validation

        Args:
            source: The source type to be converted.
            handler: The handler to get the CoreSchema.

        Returns:
            A Pydantic CoreSchema with the Duration validation.
        """
        return core_schema.no_info_wrap_validator_function(
            cls._validate,
            core_schema.timedelta_schema(),
            serialization=core_schema.plain_serializer_function_ser_schema(
                lambda instance: instance.to_iso8601_string(), when_used='json-unless-none'
            ),
        )

    def to_iso8601_string(self) -> str:
        """
        Convert a Duration object to an ISO 8601 string.

        In addition to the standard ISO 8601 format, this method also supports the representation of fractions of a second and negative durations.

        Returns:
            The ISO 8601 string representation of the duration.
        """
        # Determine if the duration is negative
        is_negative = self.total_seconds() < 0

        # Extracting components from the Duration object (use abs for negative durations)
        years = abs(self.years) if is_negative else self.years
        months = abs(self.months) if is_negative else self.months
        days = abs(self._days) if is_negative else self._days
        hours = abs(self.hours) if is_negative else self.hours
        minutes = abs(self.minutes) if is_negative else self.minutes
        seconds = abs(self.remaining_seconds) if is_negative else self.remaining_seconds
        milliseconds = abs(self.microseconds) // 1000 if is_negative else self.microseconds // 1000
        microseconds = abs(self.microseconds) % 1000 if is_negative else self.microseconds % 1000

        # Constructing the ISO 8601 duration string
        iso_duration = 'P'
        if years or months or days:
            if years:
                iso_duration += f'{years}Y'
            if months:
                iso_duration += f'{months}M'
            if days:
                iso_duration += f'{days}D'

        if hours or minutes or seconds or milliseconds or microseconds:
            iso_duration += 'T'
            if hours:
                iso_duration += f'{hours}H'
            if minutes:
                iso_duration += f'{minutes}M'
            if seconds or milliseconds or microseconds:
                iso_duration += f'{seconds}'
                if milliseconds or microseconds:
                    iso_duration += f'.{milliseconds:03d}'
                if microseconds:
                    iso_duration += f'{microseconds:03d}'
                iso_duration += 'S'

        # Prefix with '-' if the duration is negative
        if is_negative:
            iso_duration = '-' + iso_duration

        if iso_duration == 'P':
            iso_duration = 'P0D'

        return iso_duration

    @classmethod
    def _validate(cls, value: Any, handler: core_schema.ValidatorFunctionWrapHandler) -> Duration:
        """Validate the Duration object and return it.

        Args:
            value: The value to validate.
            handler: The handler to get the CoreSchema.

        Returns:
            The validated value or raises a PydanticCustomError.
        """

        if isinstance(value, _Duration):
            return Duration(
                years=value.years,
                months=value.months,
                weeks=value.weeks,
                days=value.remaining_days,
                hours=value.hours,
                minutes=value.minutes,
                seconds=value.remaining_seconds,
                microseconds=value.microseconds,
            )

        if isinstance(value, timedelta):
            return Duration(
                days=value.days,
                seconds=value.seconds,
                microseconds=value.microseconds,
            )

        assert isinstance(value, str)
        try:
            # 'P' alone is not a valid ISO 8601 duration (must have at least one designator like P0D)
            if value in ('P', '-P'):
                raise ValueError('P alone is not a valid ISO 8601 duration')

            # https://github.com/python-pendulum/pendulum/issues/532
            if value.startswith('-'):
                parsed = parse(value.lstrip('-'), exact=True)
            else:
                parsed = parse(value, exact=True)
            if not isinstance(parsed, _Duration):
                raise ValueError(f'value is not a valid duration it is a {type(parsed)}')
            if value.startswith('-'):
                parsed = -parsed

            return Duration(
                years=parsed.years,
                months=parsed.months,
                weeks=parsed.weeks,
                days=parsed.remaining_days,
                hours=parsed.hours,
                minutes=parsed.minutes,
                seconds=parsed.remaining_seconds,
                microseconds=parsed.microseconds,
            )
        except Exception as exc:
            raise PydanticCustomError('value_error', 'value is not a valid duration') from exc


class Interval(_Interval[Any]):
    """A `pendulum.Interval` object. At runtime, this type decomposes into pendulum.Interval automatically.
    This type exists because Pydantic throws a fit on unknown types.

    ```python
    from pydantic import BaseModel
    from pydantic_extra_types.pendulum_dt import Interval


    class test_model(BaseModel):
        interval: Interval


    print(test_model(interval='2021-01-01T00:00:00+00:00/2021-01-02T00:00:00+00:00'))

    # > test_model(interval=<Interval [2021-01-01 00:00:00+00:00 -> 2021-01-02 00:00:00+00:00]>)
    ```
    """

    __slots__: list[str] = []

    @classmethod
    def __get_pydantic_core_schema__(cls, source: type[Any], handler: GetCoreSchemaHandler) -> core_schema.CoreSchema:
        """Return a Pydantic CoreSchema with the Interval validation.

        Args:
            source: The source type to be converted.
            handler: The handler to get the CoreSchema.

        Returns:
            A Pydantic CoreSchema with the Interval validation.
        """
        return core_schema.no_info_wrap_validator_function(cls._validate, core_schema.str_schema())

    @classmethod
    def _instance(cls, value: _Interval[Any]) -> Interval:
        return Interval(value.start, value.end, absolute=getattr(value, '_absolute', False))

    @classmethod
    def _validate(cls, value: Any, handler: core_schema.ValidatorFunctionWrapHandler) -> Interval:
        """Validate the interval object and return it.

        Args:
            value: The value to validate.
            handler: The handler to get the CoreSchema.

        Returns:
            The validated value or raises a PydanticCustomError.
        """
        # if we are passed an existing instance, pass it straight through.
        if isinstance(value, _Interval):
            return cls._instance(value)

        # otherwise, parse it.
        try:
            parsed = parse(value)
            if isinstance(parsed, _Interval):
                return cls._instance(parsed)
            raise ValueError(f'value is not a valid interval it is a {type(parsed)}')
        except Exception as exc:
            raise PydanticCustomError('value_error', 'value is not a valid interval') from exc



"""The `pydantic_extra_types.phone_numbers` module provides the
[`PhoneNumber`][pydantic_extra_types.phone_numbers.PhoneNumber] data type.

This class depends on the [phonenumbers](https://pypi.org/project/phonenumbers/) package,
which is a Python port of Google's [libphonenumber](https://github.com/google/libphonenumber/).
"""

from __future__ import annotations

from collections.abc import Sequence
from dataclasses import dataclass
from functools import partial
from typing import Any, ClassVar

from pydantic import GetCoreSchemaHandler, GetJsonSchemaHandler
from pydantic_core import PydanticCustomError, core_schema

try:
    import phonenumbers
    from phonenumbers import PhoneNumber as BasePhoneNumber
    from phonenumbers.phonenumberutil import NumberParseException
except ModuleNotFoundError as e:  # pragma: no cover
    raise RuntimeError(
        '`PhoneNumber` requires "phonenumbers" to be installed. You can install it with "pip install phonenumbers"'
    ) from e


class PhoneNumber(str):
    """A wrapper around the `phonenumbers.PhoneNumber` object.

    It provides class-level configuration points you can change by subclassing:

    ## Examples

    ### Normal usage:

    ```python
    from pydantic import BaseModel
    from pydantic_extra_types.phone_numbers import PhoneNumber


    class Contact(BaseModel):
        name: str
        phone: PhoneNumber


    c = Contact(name='Alice', phone='+1 650-253-0000')
    print(c.phone)
    # > tel:+1-650-253-0000 (formatted using RFC3966 by default)
    ```

    ### Changing defaults by subclassing:

    ```python
    from pydantic_extra_types.phone_numbers import PhoneNumber


    class USPhone(PhoneNumber):
        default_region_code = 'US'
        supported_regions = ['US']
        phone_format = 'NATIONAL'


    # Now parsing will accept national numbers for the US
    p = USPhone('650-253-0000')
    print(p)
    # > 650-253-0000
    ```

    ### Changing defaults by using the provided validator annotation:

    ```python
    from typing import Annotated, Union
    import phonenumbers
    from pydantic import BaseModel
    from pydantic_extra_types.phone_numbers import PhoneNumberValidator

    E164NumberType = Annotated[Union[str, phonenumbers.PhoneNumber], PhoneNumberValidator(number_format='E164')]


    class Model(BaseModel):
        phone: E164NumberType


    m = Model(phone='+1 650-253-0000')
    print(m.phone)
    # > +16502530000
    ```

    """

    default_region_code: ClassVar[str | None] = None
    """The default region code to use when parsing phone numbers without an international prefix."""

    supported_regions: list[str] = []
    """The supported regions. If empty, all regions are supported."""

    phone_format: str = 'RFC3966'
    """The format of the phone number."""

    @classmethod
    def __get_pydantic_json_schema__(
        cls, schema: core_schema.CoreSchema, handler: GetJsonSchemaHandler
    ) -> dict[str, Any]:
        json_schema = handler(schema)
        json_schema.update({'format': 'phone'})
        return json_schema

    @classmethod
    def __get_pydantic_core_schema__(cls, source: type[Any], handler: GetCoreSchemaHandler) -> core_schema.CoreSchema:
        return core_schema.with_info_after_validator_function(
            cls._validate,
            core_schema.str_schema(),
        )

    @classmethod
    def _validate(cls, phone_number: str, _: core_schema.ValidationInfo) -> str:
        try:
            parsed_number = phonenumbers.parse(phone_number, cls.default_region_code)
        except phonenumbers.phonenumberutil.NumberParseException as exc:
            raise PydanticCustomError('value_error', 'value is not a valid phone number') from exc
        if not phonenumbers.is_valid_number(parsed_number):
            raise PydanticCustomError('value_error', 'value is not a valid phone number')

        if cls.supported_regions and not any(
            phonenumbers.is_valid_number_for_region(parsed_number, region_code=region)
            for region in cls.supported_regions
        ):
            raise PydanticCustomError('value_error', 'value is not from a supported region')

        return phonenumbers.format_number(parsed_number, getattr(phonenumbers.PhoneNumberFormat, cls.phone_format))

    def __eq__(self, other: Any) -> bool:
        return super().__eq__(other)

    def __hash__(self) -> int:
        return super().__hash__()


@dataclass(frozen=True)
class PhoneNumberValidator:
    """An annotation to validate `phonenumbers.PhoneNumber` objects.

    Example:
        ```python
        from typing import Annotated, Union

        import phonenumbers
        from pydantic import BaseModel
        from pydantic_extra_types.phone_numbers import PhoneNumberValidator

        MyNumberType = Annotated[Union[str, phonenumbers.PhoneNumber], PhoneNumberValidator()]

        USNumberType = Annotated[
            Union[str, phonenumbers.PhoneNumber], PhoneNumberValidator(supported_regions=['US'], default_region='US')
        ]


        class SomeModel(BaseModel):
            phone_number: MyNumberType
            us_number: USNumberType
        ```
    """

    default_region: str | None = None
    """The default region code to use when parsing phone numbers without an international prefix.

    If `None` (the default), the region must be supplied in the phone number as an international prefix.
    """

    number_format: str = 'RFC3966'
    """The format of the phone number to return. See `phonenumbers.PhoneNumberFormat` for valid values."""

    supported_regions: Sequence[str] | None = None
    """The supported regions. If empty (the default), all regions are supported."""

    def __post_init__(self) -> None:
        if self.default_region and self.default_region not in phonenumbers.SUPPORTED_REGIONS:
            raise ValueError(f'Invalid default region code: {self.default_region}')

        if self.number_format not in (
            number_format
            for number_format in dir(phonenumbers.PhoneNumberFormat)
            if not number_format.startswith('_') and number_format.isupper()
        ):
            raise ValueError(f'Invalid number format: {self.number_format}')

        if self.supported_regions:
            for supported_region in self.supported_regions:
                if supported_region not in phonenumbers.SUPPORTED_REGIONS:
                    raise ValueError(f'Invalid supported region code: {supported_region}')

    @staticmethod
    def _parse(
        region: str | None,
        number_format: str,
        supported_regions: Sequence[str] | None,
        phone_number: Any,
    ) -> str:
        if not phone_number:
            raise PydanticCustomError('value_error', 'value is not a valid phone number')

        if not isinstance(phone_number, (str, BasePhoneNumber)):
            raise PydanticCustomError('value_error', 'value is not a valid phone number')

        parsed_number = None
        if isinstance(phone_number, BasePhoneNumber):
            parsed_number = phone_number
        else:
            try:
                parsed_number = phonenumbers.parse(phone_number, region=region)
            except NumberParseException as exc:
                raise PydanticCustomError('value_error', 'value is not a valid phone number') from exc

        if not phonenumbers.is_valid_number(parsed_number):
            raise PydanticCustomError('value_error', 'value is not a valid phone number')

        if supported_regions and not any(
            phonenumbers.is_valid_number_for_region(parsed_number, region_code=region) for region in supported_regions
        ):
            raise PydanticCustomError('value_error', 'value is not from a supported region')

        return phonenumbers.format_number(parsed_number, getattr(phonenumbers.PhoneNumberFormat, number_format))

    def __get_pydantic_core_schema__(self, source: type[Any], handler: GetCoreSchemaHandler) -> core_schema.CoreSchema:
        return core_schema.no_info_before_validator_function(
            partial(
                self._parse,
                self.default_region,
                self.number_format,
                self.supported_regions,
            ),
            core_schema.str_schema(),
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls, schema: core_schema.CoreSchema, handler: GetJsonSchemaHandler
    ) -> dict[str, Any]:
        json_schema = handler(schema)
        json_schema.update({'format': 'phone'})
        return json_schema

    def __hash__(self) -> int:
        return super().__hash__()



"""The `pydantic_extra_types.cron` module provides the [`CronStr`][pydantic_extra_types.cron.CronStr] data type."""

from __future__ import annotations

from datetime import datetime
from typing import Any, ClassVar

try:
    from cron_converter import Cron
    from cron_converter.sub_modules.seeker import Seeker as CronSeeker
except ModuleNotFoundError as e:  # pragma: no cover
    raise RuntimeError(
        'The `cron` module requires "cron-converter" to be installed. You can install it with "pip install cron-converter".'
    ) from e
from pydantic import GetCoreSchemaHandler
from pydantic_core import PydanticCustomError, core_schema


class CronStr(str):
    """A cron expression validated via [`cron-converter`](https://pypi.org/project/cron-converter/).
    ## Examples
    ```python
        from pydantic import BaseModel
        from pydantic_extra_types.cron import CronStr

        class Schedule(BaseModel):
            cron: CronStr

        schedule = Schedule(cron="*/5 * * * *")
        print(schedule.cron)
        >> */5 * * * *
        print(schedule.cron.minute)
        >> */5
        print(schedule.cron.next_run)
        >> 2025-10-07T22:40:00+00:00
    ```
    """

    strip_whitespace: ClassVar[bool] = True
    """Whether to strip surrounding whitespace from the input value."""
    _component_names: ClassVar[tuple[str, ...]] = (
        'minute',
        'hour',
        'day_of_the_month',
        'month',
        'day_of_the_week',
    )
    """Expected cron expression components in the order enforced by `cron-converter`."""

    minute: str
    hour: str
    day_of_the_month: str
    month: str
    day_of_the_week: str
    cron_obj: Cron

    def __new__(cls, cron_expression: str, *, _cron: Cron | None = None) -> CronStr:
        if _cron is None:
            cron_expression, cron_obj = cls._validate(cron_expression)
        else:
            cron_obj = _cron
            cron_expression = cron_obj.to_string()

        obj = super().__new__(cls, cron_expression)
        obj._apply_cron(cron_obj)
        return obj

    def _apply_cron(self, cron_obj: Cron) -> None:
        self.cron_obj = cron_obj
        self.minute, self.hour, self.day_of_the_month, self.month, self.day_of_the_week = str(self).split()

    @classmethod
    def _validate(cls, value: Any) -> tuple[str, Cron]:
        if not isinstance(value, str):
            raise PydanticCustomError('cron_str_type', 'Cron expression must be a string')

        cron_expression = value.strip()
        if not cron_expression:
            raise PydanticCustomError('cron_str_empty', 'Cron expression must not be empty')

        parts = cron_expression.split()
        if len(parts) != len(cls._component_names):
            parts_list = ', '.join(cls._component_names)
            raise PydanticCustomError(
                'cron_str_components',
                f'Cron expression must contain {len(cls._component_names)} space separated components: {parts_list}',
            )

        try:
            cron_obj = Cron(cron_expression)
        except (TypeError, ValueError) as exc:
            raise PydanticCustomError('cron_str_invalid', str(exc)) from exc

        # `cron-converter` may normalise components (e.g. remove duplicate spaces),
        # so we reuse its canonical representation.
        return cron_obj.to_string(), cron_obj

    @classmethod
    def validate(cls, __input_value: Any, _: core_schema.ValidationInfo) -> CronStr:
        cron_expression, cron_obj = cls._validate(__input_value)
        return cls(cron_expression, _cron=cron_obj)

    @classmethod
    def __get_pydantic_core_schema__(cls, source_type: Any, handler: GetCoreSchemaHandler) -> core_schema.CoreSchema:
        return core_schema.with_info_after_validator_function(
            cls.validate,
            core_schema.str_schema(strip_whitespace=cls.strip_whitespace),
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls, schema: core_schema.CoreSchema, handler: GetCoreSchemaHandler
    ) -> dict[str, Any]:
        return dict(handler(schema))

    def schedule(self, start_date: datetime | None = None, timezone_str: str | None = None) -> CronSeeker:
        """Return the iterator produced by `cron-converter` for this expression."""
        return self.cron_obj.schedule(start_date=start_date, timezone_str=timezone_str)

    def next_after(self, start_date: datetime | None = None, timezone_str: str | None = None) -> datetime:
        """Return the first run datetime after `start_date` (or now if omitted)."""
        seeker = self.schedule(start_date=start_date, timezone_str=timezone_str)
        return seeker.next()

    @property
    def next_run(self) -> str:
        """Return the next run as an ISO formatted string (shortcut for backwards compatibility)."""
        return self.next_after().isoformat()



from __future__ import annotations

import datetime
from typing import Any, Callable

import pydantic_core.core_schema
from pydantic import GetJsonSchemaHandler
from pydantic.json_schema import JsonSchemaValue
from pydantic_core import CoreSchema, core_schema

EPOCH = datetime.datetime(1970, 1, 1, tzinfo=datetime.timezone.utc)


class _Base(datetime.datetime):
    TYPE: str = ''
    SCHEMA: pydantic_core.core_schema.CoreSchema

    @classmethod
    def __get_pydantic_json_schema__(
        cls, core_schema: core_schema.CoreSchema, handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        field_schema: dict[str, Any] = {}
        field_schema.update(type=cls.TYPE, format='date-time')
        return field_schema

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source: type[Any], handler: Callable[[Any], CoreSchema]
    ) -> core_schema.CoreSchema:
        return core_schema.with_info_after_validator_function(
            cls._validate,
            cls.SCHEMA,
            serialization=core_schema.wrap_serializer_function_ser_schema(cls._f, return_schema=cls.SCHEMA),
        )

    @classmethod
    def _validate(cls, __input_value: Any, _: Any) -> datetime.datetime:
        return EPOCH + datetime.timedelta(seconds=__input_value)

    @classmethod
    def _f(cls, value: Any, serializer: Callable[[Any], Any]) -> Any:  # pragma: no cover
        raise NotImplementedError(cls)


class Number(_Base):
    """epoch.Number parses unix timestamp as float and converts it to datetime.

    ```py
    from pydantic import BaseModel

    from pydantic_extra_types import epoch


    class LogEntry(BaseModel):
        timestamp: epoch.Number


    logentry = LogEntry(timestamp=1.1)
    print(logentry)
    # > timestamp=datetime.datetime(1970, 1, 1, 0, 0, 1, 100000, tzinfo=datetime.timezone.utc)
    ```
    """

    TYPE = 'number'
    SCHEMA = core_schema.float_schema()

    @classmethod
    def _f(cls, value: Any, serializer: Callable[[float], float]) -> float:
        ts = value.timestamp()
        return serializer(ts)


class Integer(_Base):
    """epoch.Integer parses unix timestamp as integer and converts it to datetime.

    ```
    ```py
    from pydantic import BaseModel

    from pydantic_extra_types import epoch

    class LogEntry(BaseModel):
        timestamp: epoch.Integer

    logentry = LogEntry(timestamp=1)
    print(logentry)
    #> timestamp=datetime.datetime(1970, 1, 1, 0, 0, 1, tzinfo=datetime.timezone.utc)
    ```
    """

    TYPE = 'integer'
    SCHEMA = core_schema.int_schema()

    @classmethod
    def _f(cls, value: Any, serializer: Callable[[int], int]) -> int:
        ts = value.timestamp()
        return serializer(int(ts))



"""
Validation for MongoDB ObjectId fields.

Ref: https://github.com/pydantic/pydantic-extra-types/issues/133
"""

from typing import Any

from pydantic import GetCoreSchemaHandler
from pydantic_core import core_schema

try:
    from bson import ObjectId
except ModuleNotFoundError as e:  # pragma: no cover
    raise RuntimeError(
        'The `mongo_object_id` module requires "pymongo" to be installed. You can install it with "pip install '
        'pymongo".'
    ) from e


class MongoObjectId(str):
    """MongoObjectId parses and validates MongoDB bson.ObjectId.

    ```py
    from pydantic import BaseModel

    from pydantic_extra_types.mongo_object_id import MongoObjectId


    class MongoDocument(BaseModel):
        id: MongoObjectId


    doc = MongoDocument(id='5f9f2f4b9d3c5a7b4c7e6c1d')
    print(doc)
    # > id='5f9f2f4b9d3c5a7b4c7e6c1d'
    ```

    Raises:
        PydanticCustomError: If the provided value is not a valid MongoDB ObjectId.
    """

    OBJECT_ID_LENGTH = 24

    @classmethod
    def __get_pydantic_core_schema__(cls, _: Any, __: GetCoreSchemaHandler) -> core_schema.CoreSchema:
        return core_schema.json_or_python_schema(
            json_schema=core_schema.str_schema(min_length=cls.OBJECT_ID_LENGTH, max_length=cls.OBJECT_ID_LENGTH),
            python_schema=core_schema.union_schema(
                [
                    core_schema.is_instance_schema(ObjectId),
                    core_schema.chain_schema(
                        [
                            core_schema.str_schema(min_length=cls.OBJECT_ID_LENGTH, max_length=cls.OBJECT_ID_LENGTH),
                            core_schema.no_info_plain_validator_function(cls.validate),
                        ]
                    ),
                ]
            ),
            serialization=core_schema.plain_serializer_function_ser_schema(lambda x: str(x), when_used='json'),
        )

    @classmethod
    def validate(cls, value: str) -> ObjectId:
        """Validate the MongoObjectId str is a valid ObjectId instance."""
        if not ObjectId.is_valid(value):
            raise ValueError(
                f"Invalid ObjectId {value} has to be 24 characters long and in the format '5f9f2f4b9d3c5a7b4c7e6c1d'."
            )

        return ObjectId(value)



"""Language definitions that are based on the [ISO 639-3](https://en.wikipedia.org/wiki/ISO_639-3) & [ISO 639-5](https://en.wikipedia.org/wiki/ISO_639-5)."""

from __future__ import annotations

from dataclasses import dataclass
from functools import lru_cache
from typing import Any, Union

from pydantic import GetCoreSchemaHandler, GetJsonSchemaHandler
from pydantic_core import PydanticCustomError, core_schema

try:
    import pycountry
except ModuleNotFoundError as e:  # pragma: no cover
    raise RuntimeError(
        'The `language_code` module requires "pycountry" to be installed.'
        ' You can install it with "pip install pycountry".'
    ) from e


@dataclass
class LanguageInfo:
    """LanguageInfo is a dataclass that contains the language information.

    Args:
        alpha2: The language code in the [ISO 639-1 alpha-2](https://en.wikipedia.org/wiki/ISO_639-1) format.
        alpha3: The language code in the [ISO 639-3 alpha-3](https://en.wikipedia.org/wiki/ISO_639-3) format.
        name: The language name.
    """

    alpha2: Union[str, None]
    alpha3: str
    name: str


@lru_cache
def _languages() -> list[LanguageInfo]:
    """Return a list of LanguageInfo objects containing the language information.

    Returns:
        A list of LanguageInfo objects containing the language information.
    """
    return [
        LanguageInfo(
            alpha2=getattr(language, 'alpha_2', None),
            alpha3=language.alpha_3,
            name=language.name,
        )
        for language in pycountry.languages
    ]


@lru_cache
def _index_by_alpha2() -> dict[str, LanguageInfo]:
    """Return a dictionary with the language code in the [ISO 639-1 alpha-2](https://en.wikipedia.org/wiki/ISO_639-1) format as the key and the LanguageInfo object as the value."""
    return {language.alpha2: language for language in _languages() if language.alpha2 is not None}


@lru_cache
def _index_by_alpha3() -> dict[str, LanguageInfo]:
    """Return a dictionary with the language code in the [ISO 639-3 alpha-3](https://en.wikipedia.org/wiki/ISO_639-3) format as the key and the LanguageInfo object as the value."""
    return {language.alpha3: language for language in _languages()}


@lru_cache
def _index_by_name() -> dict[str, LanguageInfo]:
    """Return a dictionary with the language name as the key and the LanguageInfo object as the value."""
    return {language.name: language for language in _languages()}


class LanguageAlpha2(str):
    """LanguageAlpha2 parses languages codes in the [ISO 639-1 alpha-2](https://en.wikipedia.org/wiki/ISO_639-1)
    format.

    ```py
    from pydantic import BaseModel

    from pydantic_extra_types.language_code import LanguageAlpha2


    class Movie(BaseModel):
        audio_lang: LanguageAlpha2
        subtitles_lang: LanguageAlpha2


    movie = Movie(audio_lang='de', subtitles_lang='fr')
    print(movie)
    # > audio_lang='de' subtitles_lang='fr'
    ```
    """

    @classmethod
    def _validate(cls, __input_value: str, _: core_schema.ValidationInfo) -> LanguageAlpha2:
        """Validate a language code in the ISO 639-1 alpha-2 format from the provided str value.

        Args:
            __input_value: The str value to be validated.
            _: The Pydantic ValidationInfo.

        Returns:
            The validated language code in the ISO 639-1 alpha-2 format.
        """
        if __input_value not in _index_by_alpha2():
            raise PydanticCustomError('language_alpha2', 'Invalid language alpha2 code')
        return cls(__input_value)

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source: type[Any], handler: GetCoreSchemaHandler
    ) -> core_schema.AfterValidatorFunctionSchema:
        """Return a Pydantic CoreSchema with the language code in the ISO 639-1 alpha-2 format validation.

        Args:
            source: The source type.
            handler: The handler to get the CoreSchema.

        Returns:
            A Pydantic CoreSchema with the language code in the ISO 639-1 alpha-2 format validation.
        """
        return core_schema.with_info_after_validator_function(
            cls._validate,
            core_schema.str_schema(to_lower=True),
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls, schema: core_schema.CoreSchema, handler: GetJsonSchemaHandler
    ) -> dict[str, Any]:
        """Return a Pydantic JSON Schema with the language code in the ISO 639-1 alpha-2 format validation.

        Args:
            schema: The Pydantic CoreSchema.
            handler: The handler to get the JSON Schema.

        Returns:
            A Pydantic JSON Schema with the language code in the ISO 639-1 alpha-2 format validation.
        """
        json_schema = handler(schema)
        json_schema.update({'pattern': r'^\w{2}$'})
        return json_schema

    @property
    def alpha3(self) -> str:
        """The language code in the [ISO 639-3 alpha-3](https://en.wikipedia.org/wiki/ISO_639-3) format."""
        return _index_by_alpha2()[self].alpha3

    @property
    def name(self) -> str:
        """The language name."""
        return _index_by_alpha2()[self].name


class LanguageName(str):
    """LanguageName parses languages names listed in the [ISO 639-3 standard](https://en.wikipedia.org/wiki/ISO_639-3)
    format.

    ```py
    from pydantic import BaseModel

    from pydantic_extra_types.language_code import LanguageName


    class Movie(BaseModel):
        audio_lang: LanguageName
        subtitles_lang: LanguageName


    movie = Movie(audio_lang='Dutch', subtitles_lang='Mandarin Chinese')
    print(movie)
    # > audio_lang='Dutch' subtitles_lang='Mandarin Chinese'
    ```
    """

    @classmethod
    def _validate(cls, __input_value: str, _: core_schema.ValidationInfo) -> LanguageName:
        """Validate a language name from the provided str value.

        Args:
            __input_value: The str value to be validated.
            _: The Pydantic ValidationInfo.

        Returns:
            The validated language name.
        """
        if __input_value not in _index_by_name():
            raise PydanticCustomError('language_name', 'Invalid language name')
        return cls(__input_value)

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source: type[Any], handler: GetCoreSchemaHandler
    ) -> core_schema.AfterValidatorFunctionSchema:
        """Return a Pydantic CoreSchema with the language name validation.

        Args:
            source: The source type.
            handler: The handler to get the CoreSchema.

        Returns:
            A Pydantic CoreSchema with the language name validation.
        """
        return core_schema.with_info_after_validator_function(
            cls._validate,
            core_schema.str_schema(),
            serialization=core_schema.to_string_ser_schema(),
        )

    @property
    def alpha2(self) -> Union[str, None]:
        """The language code in the [ISO 639-1 alpha-2](https://en.wikipedia.org/wiki/ISO_639-1) format. Does not exist for all languages."""
        return _index_by_name()[self].alpha2

    @property
    def alpha3(self) -> str:
        """The language code in the [ISO 639-3 alpha-3](https://en.wikipedia.org/wiki/ISO_639-3) format."""
        return _index_by_name()[self].alpha3


class ISO639_3(str):
    """ISO639_3 parses Language in the [ISO 639-3 alpha-3](https://en.wikipedia.org/wiki/ISO_639-3_alpha-3)
    format.

    ```py
    from pydantic import BaseModel

    from pydantic_extra_types.language_code import ISO639_3


    class Language(BaseModel):
        alpha_3: ISO639_3


    lang = Language(alpha_3='ssr')
    print(lang)
    # > alpha_3='ssr'
    ```
    """

    allowed_values_list = [lang.alpha_3 for lang in pycountry.languages]
    allowed_values = set(allowed_values_list)

    @classmethod
    def _validate(cls, __input_value: str, _: core_schema.ValidationInfo) -> ISO639_3:
        """Validate a ISO 639-3 language code from the provided str value.

        Args:
            __input_value: The str value to be validated.
            _: The Pydantic ValidationInfo.

        Returns:
            The validated ISO 639-3 language code.

        Raises:
            PydanticCustomError: If the ISO 639-3 language code is not valid.
        """
        if __input_value not in cls.allowed_values:
            raise PydanticCustomError(
                'ISO649_3', 'Invalid ISO 639-3 language code. See https://en.wikipedia.org/wiki/ISO_639-3'
            )
        return cls(__input_value)

    @classmethod
    def __get_pydantic_core_schema__(
        cls, _: type[Any], __: GetCoreSchemaHandler
    ) -> core_schema.AfterValidatorFunctionSchema:
        """Return a Pydantic CoreSchema with the ISO 639-3 language code validation.

        Args:
            _: The source type.
            __: The handler to get the CoreSchema.

        Returns:
            A Pydantic CoreSchema with the ISO 639-3 language code validation.

        """
        return core_schema.with_info_after_validator_function(
            cls._validate,
            core_schema.str_schema(min_length=3, max_length=3),
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls, schema: core_schema.CoreSchema, handler: GetJsonSchemaHandler
    ) -> dict[str, Any]:
        """Return a Pydantic JSON Schema with the ISO 639-3 language code validation.

        Args:
            schema: The Pydantic CoreSchema.
            handler: The handler to get the JSON Schema.

        Returns:
            A Pydantic JSON Schema with the ISO 639-3 language code validation.

        """
        json_schema = handler(schema)
        json_schema.update({'enum': cls.allowed_values_list})
        return json_schema


class ISO639_5(str):
    """ISO639_5 parses Language in the [ISO 639-5 alpha-3](https://en.wikipedia.org/wiki/ISO_639-5_alpha-3)
    format.

    ```py
    from pydantic import BaseModel

    from pydantic_extra_types.language_code import ISO639_5


    class Language(BaseModel):
        alpha_3: ISO639_5


    lang = Language(alpha_3='gem')
    print(lang)
    # > alpha_3='gem'
    ```
    """

    allowed_values_list = [lang.alpha_3 for lang in pycountry.language_families]
    allowed_values_list.sort()
    allowed_values = set(allowed_values_list)

    @classmethod
    def _validate(cls, __input_value: str, _: core_schema.ValidationInfo) -> ISO639_5:
        """Validate a ISO 639-5 language code from the provided str value.

        Args:
            __input_value: The str value to be validated.
            _: The Pydantic ValidationInfo.

        Returns:
            The validated ISO 639-3 language code.

        Raises:
            PydanticCustomError: If the ISO 639-5 language code is not valid.
        """
        if __input_value not in cls.allowed_values:
            raise PydanticCustomError(
                'ISO649_5', 'Invalid ISO 639-5 language code. See https://en.wikipedia.org/wiki/ISO_639-5'
            )
        return cls(__input_value)

    @classmethod
    def __get_pydantic_core_schema__(
        cls, _: type[Any], __: GetCoreSchemaHandler
    ) -> core_schema.AfterValidatorFunctionSchema:
        """Return a Pydantic CoreSchema with the ISO 639-5 language code validation.

        Args:
            _: The source type.
            __: The handler to get the CoreSchema.

        Returns:
            A Pydantic CoreSchema with the ISO 639-5 language code validation.

        """
        return core_schema.with_info_after_validator_function(
            cls._validate,
            core_schema.str_schema(min_length=3, max_length=3),
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls, schema: core_schema.CoreSchema, handler: GetJsonSchemaHandler
    ) -> dict[str, Any]:
        """Return a Pydantic JSON Schema with the ISO 639-5 language code validation.

        Args:
            schema: The Pydantic CoreSchema.
            handler: The handler to get the JSON Schema.

        Returns:
            A Pydantic JSON Schema with the ISO 639-5 language code validation.

        """
        json_schema = handler(schema)
        json_schema.update({'enum': cls.allowed_values_list})
        return json_schema



from __future__ import annotations

import typing
from dataclasses import dataclass
from pathlib import Path
from typing import Annotated

import pydantic
from pydantic.types import PathType
from pydantic_core import core_schema

ExistingPath = typing.Union[pydantic.FilePath, pydantic.DirectoryPath]


@dataclass
class ResolvedPathType(PathType):
    """A custom PathType that resolves the path to its absolute form.

    Args:
        path_type (typing.Literal['file', 'dir', 'new']): The type of path to resolve. Can be 'file', 'dir' or 'new'.

    Returns:
        Resolved path as a pathlib.Path object.

    Example:
        ```python
        from pydantic import BaseModel
        from pydantic_extra_types.path import ResolvedFilePath, ResolvedDirectoryPath, ResolvedNewPath


        class MyModel(BaseModel):
            file_path: ResolvedFilePath
            dir_path: ResolvedDirectoryPath
            new_path: ResolvedNewPath


        model = MyModel(file_path='~/myfile.txt', dir_path='~/mydir', new_path='~/newfile.txt')
        print(model.file_path)
        # > file_path=PosixPath('/home/user/myfile.txt') dir_path=PosixPath('/home/user/mydir') new_path=PosixPath('/home/user/newfile.txt')"""

    @staticmethod
    def validate_file(path: Path, _: core_schema.ValidationInfo) -> Path:
        return PathType.validate_file(path.expanduser().resolve(), _)

    @staticmethod
    def validate_directory(path: Path, _: core_schema.ValidationInfo) -> Path:
        return PathType.validate_directory(path.expanduser().resolve(), _)

    @staticmethod
    def validate_new(path: Path, _: core_schema.ValidationInfo) -> Path:
        return PathType.validate_new(path.expanduser().resolve(), _)

    def __hash__(self) -> int:
        return hash(type(self.path_type))


ResolvedFilePath = Annotated[Path, ResolvedPathType('file')]
ResolvedDirectoryPath = Annotated[Path, ResolvedPathType('dir')]
ResolvedNewPath = Annotated[Path, ResolvedPathType('new')]
ResolvedExistingPath = typing.Union[ResolvedFilePath, ResolvedDirectoryPath]



"""SemanticVersion definition that is based on the Semantiv Versioning Specification [semver](https://semver.org/)."""

from typing import Any, Callable

from pydantic import GetJsonSchemaHandler
from pydantic.json_schema import JsonSchemaValue
from pydantic_core import core_schema

try:
    import semver
except ModuleNotFoundError as e:  # pragma: no cover
    raise RuntimeError(
        'The `semantic_version` module requires "semver" to be installed. You can install it with "pip install semver".'
    ) from e


class SemanticVersion(semver.Version):
    """Semantic version based on the official [semver thread](https://python-semver.readthedocs.io/en/latest/advanced/combine-pydantic-and-semver.html)."""

    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        _source_type: Any,
        _handler: Callable[[Any], core_schema.CoreSchema],
    ) -> core_schema.CoreSchema:
        def validate_from_str(value: str) -> SemanticVersion:
            return cls.parse(value)

        from_str_schema = core_schema.chain_schema(
            [
                core_schema.str_schema(),
                core_schema.no_info_plain_validator_function(validate_from_str),
            ]
        )

        return core_schema.json_or_python_schema(
            json_schema=from_str_schema,
            python_schema=core_schema.union_schema(
                [
                    core_schema.is_instance_schema(semver.Version),
                    from_str_schema,
                ]
            ),
            serialization=core_schema.to_string_ser_schema(),
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls, _core_schema: core_schema.CoreSchema, handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        return handler(
            core_schema.str_schema(
                pattern=r'^(0|[1-9]\d*)\.(0|[1-9]\d*)\.(0|[1-9]\d*)(?:-((?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)(?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*))?(?:\+([0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?$'
            )
        )

    @classmethod
    def validate_from_str(cls, value: str) -> 'SemanticVersion':
        return cls.parse(value)
